# From https://github.com/vsitzmann/siren/blob/master/explore_siren.ipynb
from torch import nn
import torch
import numpy as np

from .abs_layer import AbsLayer


class SineLayer(nn.Module):
    """Sinuosidal layer for the SIREN model."""

    # See paper sec. 3.2, final paragraph, and supplement Sec. 1.5 for discussion of omega_0.

    # If is_first=True, omega_0 is a frequency factor which simply multiplies the activations before the
    # nonlinearity. Different signals may require different omega_0 in the first layer - this is a
    # hyperparameter.

    # If is_first=False, then the weights will be divided by omega_0 so as to keep the magnitude of
    # activations constant, but boost gradients to the weight matrix (see supplement Sec. 1.5)

    def __init__(
        self, in_features, out_features, bias=True, is_first=False, omega_0=30
    ):
        super().__init__()
        self.omega_0 = omega_0
        self.is_first = is_first

        self.in_features = in_features
        self.linear = nn.Linear(in_features, out_features, bias=bias)

        self.init_weights()

    def init_weights(self):
        with torch.no_grad():
            if self.is_first:
                self.linear.weight.uniform_(-1 / self.in_features, 1 / self.in_features)
            else:
                self.linear.weight.uniform_(
                    -np.sqrt(6 / self.in_features) / self.omega_0,
                    np.sqrt(6 / self.in_features) / self.omega_0,
                )

    def forward(self, input):
        return torch.sin(self.omega_0 * self.linear(input))

    def forward_with_intermediate(self, input):
        # For visualization of activation distributions
        intermediate = self.omega_0 * self.linear(input)
        return torch.sin(intermediate), intermediate


class Siren(nn.Module):
    """SIREN model as described in the paper by Sitzmann et al. (2020)."""

    def __init__(
        self,
        in_features,
        hidden_features,
        hidden_layers,
        out_features,
        outermost_linear=False,
        outermost_activation=AbsLayer(),
        first_omega_0=30,
        hidden_omega_0=30.0,
    ):

        super().__init__()

        self.in_features = in_features
        self.net = []
        self.net.append(
            SineLayer(
                in_features, hidden_features, is_first=True, omega_0=first_omega_0
            )
        )

        for i in range(hidden_layers):
            self.net.append(
                SineLayer(
                    hidden_features,
                    hidden_features,
                    is_first=False,
                    omega_0=hidden_omega_0,
                )
            )

        if outermost_linear:
            final_linear = nn.Linear(hidden_features, out_features)

            with torch.no_grad():
                final_linear.weight.uniform_(
                    -np.sqrt(6 / hidden_features) / hidden_omega_0,
                    np.sqrt(6 / hidden_features) / hidden_omega_0,
                )

            self.net.append(final_linear)
            self.net.append(outermost_activation)
        else:
            self.net.append(
                SineLayer(
                    hidden_features,
                    out_features,
                    is_first=False,
                    omega_0=hidden_omega_0,
                )
            )

        self.net = nn.Sequential(*self.net)

    def forward(self, coords):
        coords = (
            coords.clone().detach().requires_grad_(True)
        )  # allows to take derivative w.r.t. input
        output = self.net(coords)
        return output

    def forward_with_activations(self, coords, retain_grad=False):
        """Returns not only model output, but also intermediate activations.
        Only used for visualizing activations later!"""
        activations = OrderedDict()

        activation_count = 0
        x = coords.clone().detach().requires_grad_(True)
        activations["input"] = x
        for i, layer in enumerate(self.net):
            if isinstance(layer, SineLayer):
                x, intermed = layer.forward_with_intermediate(x)

                if retain_grad:
                    x.retain_grad()
                    intermed.retain_grad()

                activations[
                    "_".join((str(layer.__class__), "%d" % activation_count))
                ] = intermed
                activation_count += 1
            else:
                x = layer(x)

                if retain_grad:
                    x.retain_grad()

            activations["_".join((str(layer.__class__), "%d" % activation_count))] = x
            activation_count += 1

        return activations



import torch

from nidn.utils.global_constants import PI
from ..fdtd_integration.calculate_transmission_reflection_coefficients import (
    calculate_transmission_reflection_coefficients,
)
from ..utils.load_default_cfg import load_default_cfg


def test_calculate_coefficient():
    cfg = load_default_cfg()
    time_points = torch.tensor([0.002 * PI * i for i in range(1000)])
    signal_a = 2 * torch.sin(time_points)
    signal_b = torch.sin(time_points)
    signal_array = [signal_a, signal_b]
    (
        transmission_coefficient_ms,
        reflection_coefficient_ms,
    ) = calculate_transmission_reflection_coefficients(signal_array, signal_array)
    assert transmission_coefficient_ms.item() - 0.25 < 1e-7
    assert reflection_coefficient_ms.item() - 0.25 < 1e-7


if __name__ == "__main__":
    test_calculate_coefficient()



import torch

from ..materials.material_collection import MaterialCollection
from ..training.model.model_to_eps_grid import model_to_eps_grid
from ..training.model.init_network import init_network
from ..trcwa.compute_target_frequencies import compute_target_frequencies
from ..utils.load_default_cfg import load_default_cfg


def test_model_to_eps_grid():
    """Tests if these functions run. The correctness of the results is not checked."""

    # Test classification
    cfg = load_default_cfg()
    cfg.type = "classification"

    # Determine target frequencies
    cfg.target_frequencies = compute_target_frequencies(
        cfg.physical_wavelength_range[0],
        cfg.physical_wavelength_range[1],
        cfg.N_freq,
        cfg.freq_distribution,
    )

    cfg.material_collection = MaterialCollection(cfg.target_frequencies)
    cfg.N_materials = cfg.material_collection.N_materials
    # If classification, the model outputs a likelihood for the presence of each material.
    cfg.out_features = cfg.N_materials

    model = init_network(cfg)
    eps, ids = model_to_eps_grid(model, cfg)

    assert ids is not None
    assert eps.shape == torch.Size([cfg.Nx, cfg.Ny, cfg.N_layers, cfg.N_freq])

    # Test regression
    cfg.out_features = 2

    cfg.type = "regression"

    model = init_network(cfg)
    eps, ids = model_to_eps_grid(model, cfg)
    assert ids is None
    assert eps.shape == torch.Size([cfg.Nx, cfg.Ny, cfg.N_layers, cfg.N_freq])



import torch
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.ticker import AutoLocator
from ..utils.convert_units import freq_to_wl, wl_to_phys_wl
from ..utils.compute_spectrum import compute_spectrum
from ..training.model.model_to_eps_grid import model_to_eps_grid
from ..utils.global_constants import NIDN_PLOT_COLOR_1, NIDN_PLOT_COLOR_2, NIDN_FONTSIZE


def _add_plot(
    fig,
    target_frequencies,
    produced_spectrum,
    target_spectrum,
    ylimits,
    nr,
    type_name,
    logscale=False,
    markers=True,
):
    fontsize = NIDN_FONTSIZE
    markersize = 8 if markers else 0

    ax = fig.add_subplot(nr)
    freqs = wl_to_phys_wl(freq_to_wl(target_frequencies)) * 1e6  # in µm
    ax.plot(
        freqs,
        target_spectrum,
        marker="o",
        c=NIDN_PLOT_COLOR_2,
        lw=4,
        markersize=markersize,
    )
    ax.plot(
        freqs,
        produced_spectrum,
        linestyle=("--" if not markers else "-"),
        marker="o",
        c=NIDN_PLOT_COLOR_1,
        lw=4,
        markersize=markersize,
    )
    ax.legend(
        [f"Target {type_name}", f"Produced {type_name}"],
        # loc="lower center",
        fontsize=fontsize,
    )
    ax.set_xlabel("Wavelength [µm]", fontsize=fontsize + 2)
    ax.set_ylabel(f"{type_name}", fontsize=fontsize + 2)
    if logscale:
        ax.set_xscale("log")
    ax.tick_params(axis="both", which="major", labelsize=fontsize)

    ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, pos: "{:.1f}".format(x)))
    ax.xaxis.set_major_locator(AutoLocator())
    plt.minorticks_off()
    # ax.xaxis.set_minor_formatter(FormatStrFormatter("%.1f"))
    # ax.axhspan(-6, 0, facecolor="gray", alpha=0.3)
    # ax.axhspan(1, 5, facecolor="gray", alpha=0.3)
    ax.set_ylim(ylimits)

    L1_err = abs(target_spectrum - produced_spectrum).mean()
    ax.text(
        0.01,
        1.05,
        f"L1 Error = {L1_err:.4f}",
        va="center",
        fontsize=fontsize,
        transform=ax.transAxes,
    )
    return fig


def plot_spectra(
    run_cfg,
    save_path=None,
    prod_R_spectrum=None,
    prod_T_spectrum=None,
    markers=True,
    filename=None,
):
    """Plots the produced RTA spectra together with the target spectra. Optionally saves it.

    Args:
        run_cfg (dict): The run configuration.
        save_path (str, optional): Folder to save the plot in. Defaults to None, then the plot will not be saved.
        prod_R_spectrum (torch.tensor, optional): The produced reflection spectrum. Defaults to None, then will compute from model.
        prod_T_spectrum (torch.tensor, optional): The produced transmission spectrum. Defaults to None, then will compute from model.
        markers (bool): Whether to plot markers for the target and produced spectra.
        filename (str, optional): Filename to save the plot in. Defaults to None, then the plot will be saved with the name "spectra.png".
    """
    target_R_spectrum = run_cfg.target_reflectance_spectrum
    target_T_spectrum = run_cfg.target_transmittance_spectrum

    if prod_R_spectrum is None or prod_T_spectrum is None:
        # Create epsilon grid from the model
        eps, _ = model_to_eps_grid(run_cfg.model, run_cfg)

        # Compute the spectra for the given epsilon values
        prod_R_spectrum, prod_T_spectrum = compute_spectrum(eps, run_cfg)

    target_frequencies = run_cfg.target_frequencies

    # Convert the spectra to numpy arrays for matplotlib
    prod_R_spectrum = torch.tensor(prod_R_spectrum).detach().cpu().numpy()
    prod_T_spectrum = torch.tensor(prod_T_spectrum).detach().cpu().numpy()
    target_R_spectrum = torch.tensor(target_R_spectrum).detach().cpu().numpy()
    target_T_spectrum = torch.tensor(target_T_spectrum).detach().cpu().numpy()

    # Compute absorption spectra
    prod_A_spectrum = np.ones_like(np.asarray(prod_R_spectrum)) - (
        np.asarray(prod_T_spectrum) + np.asarray(prod_R_spectrum)
    )
    target_A_spectrum = np.ones_like(np.asarray(target_R_spectrum)) - (
        np.asarray(target_T_spectrum) + np.asarray(target_R_spectrum)
    )

    # To align all plots
    ylimits = [0.0, 1.0]
    # Code below is for gray bars in spectra
    # ylimits = [-0.2, 1.075]
    # if (
    #     (max(prod_A_spectrum) > 1 or min(prod_A_spectrum) < 0)
    #     or (max(prod_T_spectrum) > 1 or min(prod_T_spectrum) < 0)
    #     or (max(prod_R_spectrum) > 1 or min(prod_R_spectrum) < 0)
    # ):
    #     ylimits = [
    #         min(min(prod_A_spectrum), min(prod_T_spectrum), min(prod_R_spectrum))
    #         + ylimits[0],
    #         max(max(prod_A_spectrum), max(prod_T_spectrum), max(prod_R_spectrum)) + 0.1,
    #     ]

    fig = plt.figure(figsize=(15, 5), dpi=300)
    fig.patch.set_facecolor("white")

    fig = _add_plot(
        fig,
        target_frequencies,
        prod_R_spectrum,
        target_R_spectrum,
        ylimits,
        121,
        "Reflectance",
        markers=markers,
    )
    fig = _add_plot(
        fig,
        target_frequencies,
        prod_T_spectrum,
        target_T_spectrum,
        ylimits,
        122,
        "Transmittance",
        markers=markers,
    )
    # fig = _add_plot(
    #     fig,
    #     target_frequencies,
    #     prod_A_spectrum,
    #     target_A_spectrum,
    #     ylimits,
    #     133,
    #     "Absorptance",
    # )

    fig.autofmt_xdate()
    plt.tight_layout()

    if save_path is not None:
        if filename is None:
            plt.savefig(save_path + "/spectra_comparison.png", dpi=150)
        else:
            plt.savefig(save_path + "/" + filename + ".png", dpi=300)



import pickle as pk
from datetime import datetime
from loguru import logger

from pathlib import Path


def save_run(run_cfg, subfolder=""):
    """Saves results of a run to a file.

    Args:
        run_cfg (DotMap): Run config.
        subfolder (str): Subfolder to save to. Defaults to "".
    """

    # datetime object containing current date and time
    now = datetime.now()
    dt_string = now.strftime("%d-%m-%Y %H-%M-%S")

    # Create subfolder if it does not exist
    Path("../results/" + subfolder + "/").mkdir(parents=True, exist_ok=True)

    filename = "../results/" + subfolder + "/" + run_cfg.name + "_" + dt_string + "_run.pk"

    logger.info("Saving run to file: {}".format(filename))
    with open(filename, "wb") as handle:
        pk.dump(run_cfg, handle, protocol=pk.HIGHEST_PROTOCOL)



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

setup(
    name="nidn",
    version="0.1.0",
    description="A package for inverse material design of nanostructures using neural networks.",
    # long_description=open("README.md").read(),
    # long_description_content_type="text/markdown",
    url="https://github.com/esa/nidn",
    author="ESA Advanced Concepts Team",
    author_email="pablo.gomez@esa.int",
    include_package_data=True,
    install_requires=[
        "dotmap>=1.3.24",
        "loguru>=0.5.3",
        "matplotlib>=3.3.3",
        "numpy>=1.20.0",
        "pandas>=1.3.1",
        "scipy>=1.6.0",
        "tqdm>=4.56.1",
        "toml>=0.10.2",
        "torch>=1.9",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.8",
    ],
    packages=[
        "nidn",
        "nidn.fdtd",
        "nidn.fdtd_integration",
        "nidn.materials",
        "nidn.plots",
        "nidn.tests",
        "nidn.training",
        "nidn.training.losses",
        "nidn.training.model",
        "nidn.training.utils",
        "nidn.trcwa",
        "nidn.trcwa.utils",
        "nidn.utils",
    ],
    python_requires=">=3.8, <4",
    project_urls={
        "Source": "https://github.com/esa/nidn/",
    },
)



import torch

# Elemental charge
E_CHARGE = 1.602176634 * 10 ** (-19)
# Mass of electron
M_E = 9.1093837015 * 10 ** (-31)
# Vacuum permittivity
EPS_0 = 8.8541878128 * 10 ** (-12)
# There is no torch.pi so we define it here
PI = torch.tensor(3.14159265358979323846264338327950288)
# Speed of light in vacuum
SPEED_OF_LIGHT = 299792458.0
# Unit size in simulations
UNIT_MAGNITUDE = 10 ** (-6)
# Vacuum permeability
MU_0 = 4e-7 * PI
# Vacuum impedance
ETA_0 = MU_0 * SPEED_OF_LIGHT


##### PLOTTING
NIDN_FONTSIZE = 16
NIDN_PLOT_COLOR_1 = "#0173b2"
NIDN_PLOT_COLOR_2 = "#de8f05"



import numpy as np
import torch

from .material_collection import MaterialCollection


class LayerBuilder:
    def __init__(self, run_cfg) -> None:
        """Initializes the builder.

        Args:
            run_cfg (DotMap): Run configuration.
        """
        self.run_cfg = run_cfg
        self.material_collection = MaterialCollection(self.run_cfg.target_frequencies)

    def _setup_grid_and_materials(
        self,
    ):
        """Utility function to setup essentials."""
        # Get the grid ticks
        x0 = torch.linspace(0, 1, self.run_cfg.Nx)
        y0 = torch.linspace(0, 1, self.run_cfg.Ny)
        # Create a meshgrid from the grid ticks
        x, y = torch.meshgrid(x0, y0)

        eps = torch.zeros(
            self.run_cfg.Nx, self.run_cfg.Ny, self.run_cfg.N_freq, dtype=torch.cdouble
        )

        return eps, x, y

    def build_uniform_layer(self, grid_material):
        """Generates a uniform layer.

        Args:
            grid_material (str): The name of the material.

        Returns:
            torch.tensor: Tensor with epsilon values for the layer
        """
        eps, _, _ = self._setup_grid_and_materials()

        eps[:, :] = self.material_collection[grid_material]

        return eps

    def build_squared_layer(self, grid_material, square_material, a=0.33):
        """Generates a square layer with a square hole in the middle.

        Args:
            grid_material (str): Name of the grid material.
            square_material (str): Name of the hole material.
            a (float, optional): Size of the square hole (0 to 0.5). Defaults to 0.33.

        Returns:
            torch.tensor: Tensor with epsilon values for the layer.
        """
        eps, grid_x, grid_y = self._setup_grid_and_materials()

        eps[:, :] = self.material_collection[grid_material]

        # Define the square
        ind = torch.logical_and(
            torch.logical_and((grid_x > a), (grid_x < 1 - a)),
            torch.logical_and((grid_y > a), (grid_y < 1 - a)),
        )

        eps[ind] = self.material_collection[square_material]

        return eps

    def build_circle_layer(
        self, grid_material, circle_material, radius=0.33, x_center=0.5, y_center=0.5
    ):
        """Generates a square layer with a circular hole in the middle.

        Args:
            grid_material (str): Name of the grid material.
            circle_material (str): Name of the hole material.
            a (float, optional): Radius of the circle (0 to 0.5). Defaults to 0.33.
            x_center (float, optional): Center of the circle in the x-direction. Defaults to 0.5.
            y_center (float, optional): Center of the circle in the y-direction. Defaults to 0.5.

        Returns:
            torch.tensor: Tensor with epsilon values for the layer
        """
        eps, grid_x, grid_y = self._setup_grid_and_materials()

        eps[:, :] = self.material_collection[grid_material]

        # Define the circle
        ind = (grid_x - x_center) ** 2 + (grid_y - y_center) ** 2 < radius**2

        eps[ind] = self.material_collection[circle_material]

        return eps



from .material_collection import MaterialCollection

import torch


def _find_closest_material(eps, run_cfg, material_collection):
    """Finds the closest matching material and distance to that (in epsilon)
    from a given epsilon grid.

    Args:
        eps (torch.tensor): A tensor of epsilon values (Nx,Ny,N_layers,N_freq).
        run_cfg (DotMap): Run configuration.
        material_collection (MaterialCollection): The material collection.

    Returns:
        tuple: The closest materials and distances to that.
    """
    Nx, Ny, N_layers, N_freq = run_cfg.Nx, run_cfg.Ny, run_cfg.N_layers, run_cfg.N_freq

    comparisons = torch.zeros(
        [
            Nx,
            Ny,
            N_layers,
            N_freq,
            material_collection.N_materials,
        ]
    )

    # Compute differences for all materials
    for idx, material_name in enumerate(material_collection.material_names):
        material_eps = material_collection[material_name]

        comparisons[..., idx] = torch.abs(eps - material_eps)

    # Find minimal entries
    comparisons = comparisons.mean(dim=3)  # Average over frequencies
    minimal_comparisons, indices = torch.min(comparisons, dim=-1)

    return minimal_comparisons, indices



from ..utils.load_default_cfg import load_default_cfg
from ..utils.validate_config import _validate_config


def test_default_config():
    """Tests whether the default config is valid"""
    cfg = load_default_cfg()
    _validate_config(cfg)


if __name__ == "__main__":
    test_default_config()



import torch
from loguru import logger


def calculate_transmission_reflection_coefficients(
    transmission_signals,
    reflection_signals,
):
    """Calculates the transmission coefficient and reflection coefficient for the signals presented.

    Args:
        transmission_signals (tuple[array,array]): Transmission signal from a free-space fdtd simulaiton, and transmission signal from a fdtd simulation with an added object
        reflection_signals (_type_): Reflection signal from a free-space fdtd simulaiton, and reflection signal from a fdtd simulation with an added object

    Returns:
        tuple[float, float]: Transmission coefficient and reflection coefficient
    """
    # Substract the free_space reflection signal from the material reflection signal, to eliminate unreflected signal from detector
    # The detector detects signal passing through both ways, and is placed between the source and the material.
    # Thus, most of the signal present is the unreflected signal, which must be removed.
    true_reflection = reflection_signals[1] - reflection_signals[0]

    _check_for_all_zero_signal(transmission_signals)
    _check_for_all_zero_signal(reflection_signals)

    # find peaks for all signals
    peaks_transmission_freespace = _torch_find_peaks(transmission_signals[0])
    peaks_transmission_material = _torch_find_peaks(transmission_signals[1])
    peaks_reflection_freespace = _torch_find_peaks(reflection_signals[0])
    peaks_reflection_material = _torch_find_peaks(true_reflection)
    transmission_coefficient = torch.tensor(0.0)

    if len(peaks_transmission_material) > 1:
        mean_squared_transmission_material = _mean_square(
            transmission_signals[1][
                peaks_transmission_material[0]
                .item() : peaks_transmission_material[-1]
                .item()
            ]
        )

    else:
        mean_squared_transmission_material = (max(transmission_signals[1]) ** 2) / 2
        logger.warning(
            "There is not enough timesteps for the transmission signal to have the proper lenght/ or no signal is transmited. The signal should at least contain 2 peaks, but {} is found.The FDTD_niter should be increased, to be sure that the resutls are valid.".format(
                len(peaks_transmission_material)
            )
        )
    mean_squared_transmission_free_space = 1
    if len(peaks_transmission_freespace) > 1:
        mean_squared_transmission_free_space = _mean_square(
            transmission_signals[0][
                peaks_transmission_freespace[0]
                .item() : peaks_transmission_freespace[-1]
                .item()
            ]
        )
    else:
        mean_squared_transmission_free_space = (max(transmission_signals[0]) ** 2) / 2

    transmission_coefficient = (
        mean_squared_transmission_material / mean_squared_transmission_free_space
    )

    reflection_coefficient = torch.tensor(0.0)

    if len(peaks_reflection_material) > 1:
        mean_squared_reflection_material = _mean_square(
            true_reflection[
                peaks_reflection_material[0]
                .item() : peaks_reflection_material[-1]
                .item()
            ]
        )
    else:
        mean_squared_reflection_material = (max(true_reflection) ** 2) / 2
        logger.warning(
            "There is not enough timesteps for the reflected signal to have the proper lenght. The signal should at least contain 2 peaks, but {} is found. The FDTD_niter should be increased, to be sure that the resutls are valid.".format(
                len(peaks_reflection_material)
            )
        )
    mean_squared_reflection_free_space = 1
    if len(peaks_reflection_freespace) > 1:
        mean_squared_reflection_free_space = _mean_square(
            reflection_signals[0][
                peaks_reflection_freespace[0]
                .item() : peaks_reflection_freespace[-1]
                .item()
            ]
        )
    else:
        mean_squared_reflection_free_space = (max(reflection_signals[0]) ** 2) / 2

    reflection_coefficient = (
        mean_squared_reflection_material / mean_squared_reflection_free_space
    )

    if transmission_coefficient < 0 or transmission_coefficient > 1:
        raise ValueError(
            f"The transmission coefficient is outside of the physical range between 0 and 1"
        )

    if reflection_coefficient < 0 or reflection_coefficient > 1:
        raise ValueError(
            f"The reflection coefficient is outside of the physical range between 0 and 1"
        )
    if transmission_coefficient + reflection_coefficient > 1:
        logger.warning(
            f"The sum of the transmission and reflection coefficient is greater than 1, which is physically impossible"
        )
    return transmission_coefficient, reflection_coefficient


def _mean_square(tensor):
    """Calculates the mean of the squared signal
    Args:
        tensor (tensor): signal to perform the calculations on

    Returns:
        torch.float: The mean square value
    """
    return torch.sum(torch.square(tensor)) / len(tensor)


def _check_for_all_zero_signal(signals):

    if _mean_square(signals[0]) <= 1e-15:
        raise ValueError(
            "The free-space signal is all zero. Increase the number of FDTD_niter to ensure that the signal reaches the detector."
        )


# From : https://stackoverflow.com/questions/54498775/pytorch-argrelmax-function-or-c
def _torch_find_peaks(signal):
    signaltemp1 = signal[1:-1] - signal[:-2]
    signaltemp2 = signal[1:-1] - signal[2:]

    # and checking where both shifts are positive;
    out1 = torch.where(signaltemp1 > 0, signaltemp1 * 0 + 1, signaltemp1 * 0)
    out2 = torch.where(signaltemp2 > 0, out1, signaltemp2 * 0)

    # argrelmax containing all peaks
    return torch.nonzero(out2, out=None) + 1



import torch


def _likelihood_regularization_loss_fn(material_predictions, L=0.5, use_max=False):
    """Computes a likelihood regularization loss for the distance between predictions
    and one / zero respectively.

    Args:
        material_predictions (torch.tensor): Predictions of the material for each grid point.
        L (float, optional): L (float, optional): Exponent of the loss, e.g. L=2 for squared loss values. Defaults to 0.5.
        use_max (bool, optional): If True, the max of the material predictions is used. Defaults to False.

    Returns:
        torch.tensor: Likelihood regularization loss
    """
    if use_max:
        # Take maximum value per grid point as distance
        max_per_point, _ = torch.max(material_predictions, dim=-1)
        distances = 1.0 - max_per_point
    else:
        # Average of minimum of distance from one and zero
        distance_to_one = torch.pow(1 - material_predictions, L)
        distance_to_zero = torch.pow(material_predictions, L)
        distances = torch.minimum(distance_to_one, distance_to_zero)

    return torch.mean(distances)



from dotmap import DotMap
from loguru import logger
import torch.nn as nn

from ..utils.nerf import NERF
from ..utils.siren import Siren


def init_network(
    run_cfg: DotMap,
):
    """Network architecture. Note that the dimensionality of the first linear layer must match the output of the encoding chosen in the config file.
    Returns:
        torch model: Initialized model
    """
    logger.debug("Initializing model..." + run_cfg.model_type)
    if run_cfg.model_type == "nerf":
        return NERF(
            in_features=run_cfg.encoding_dim,
            out_features=run_cfg.out_features,
            n_neurons=run_cfg.n_neurons,
            activation=nn.Sigmoid(),
            skip=[4],
            hidden_layers=run_cfg.hidden_layers,
        )
    elif run_cfg.model_type == "siren":
        return Siren(
            in_features=run_cfg.encoding_dim,
            out_features=run_cfg.out_features,
            hidden_features=run_cfg.n_neurons,
            hidden_layers=run_cfg.hidden_layers,
            outermost_linear=True,
            outermost_activation=nn.Sigmoid(),
            first_omega_0=run_cfg.siren_omega,
            hidden_omega_0=run_cfg.siren_omega,
        )
